{*******************************************************************************}
{                                                                               }
{ FileName     : ULife.pas                                                      }
{                                                                               }
{ Author       : Ian Lane (Email: lanei@ideal.net.au)                           }
{                                                                               }
{ Synopsis     : A simple component which implements the old computer           }
{                simulation of Life. Useful for about boxes, screen savers or   }
{                even as the core of a "Life" application.                      }
{                                                                               }
{ Distribution : This component is free for public use and components may be    }
{                freely descended from it as long as credit is given to the     }
{                author.                                                        }
{                                                                               }
{                          Copyright (c) 1998 Ian Lane                          }
{                                                                               }
{*******************************************************************************}
unit ULife;

interface

uses
  Classes, Controls, Graphics, SysUtils,
{$IFDEF VER80}
  Menus, WinProcs, WinTypes;
{$ELSE}
  Windows;
{$ENDIF}

const
  MinGridHeight = 10;
{$IFDEF VER80}
  MaxGridHeight = 250;
{$ELSE}
  MaxGridHeight = 46000;
{$ENDIF}
  DefaultGridHeight = MinGridHeight;
  MinGridWidth = 10;
{$IFDEF VER80}
  MaxGridWidth = 250;
{$ELSE}
  MaxGridWidth = 46000;
{$ENDIF}
  DefaultGridWidth = MinGridWidth;
  DefaultCellColor = clWindowText;
  DefaultGridLineColor = clWindowText;
  DefaultGridLineStyle = psDot;

type
  TPermissableGridHeight = MinGridHeight..MaxGridHeight;
  TPermissableGridWidth = MinGridWidth..MaxGridWidth;
  TGridHeightRange = 0..MaxGridHeight - 1;
  TGridWidthRange = 0..MaxGridWidth - 1;
  TGrid = Array[0..MaxGridHeight * MaxGridWidth - 1] of Byte;
  PGrid = ^TGrid;
  TLife = class(TCustomControl)
  private
    FAcceptMouseClicks : Boolean;
    FCellColor : TColor;
    FGeneration : Cardinal;
    FGrid : PGrid;
    FGridHeight : TPermissableGridHeight;
    FGridLineColor : TColor;
    FGridLineStyle : TPenStyle;
{$IFDEF VER80}
    FGridSize : Cardinal;
{$ENDIF}
    FGridWidth : TPermissableGridWidth;
    FOnChange : TNotifyEvent;
    FShowGridLines : Boolean;
    function CellCoordToGridOffset(const X : TGridWidthRange;
                                   const Y : TGridHeightRange) : Cardinal;
    function GetCells(const X : TGridWidthRange; const Y : TGridHeightRange) : Boolean;
    function GetLiveCells : Integer;
    function GetVersion : String;
    procedure SetAcceptMouseClicks(const Value : Boolean);
    procedure SetCellColor(Const Value : TColor);
    procedure SetCell(const X : TGridWidthRange; const Y : TGridHeightRange;
                      const Value : Boolean);
    procedure SetCells(const X : TGridWidthRange; const Y : TGridHeightRange;
                       const Value : Boolean);
    procedure SetGridHeight(const NewGridHeight : TPermissableGridHeight);
    procedure SetGridLineColor(Const Value : TColor);
    procedure SetGridLineStyle(Const Value : TPenStyle);
    procedure SetGridWidth(const NewGridWidth : TPermissableGridWidth);
    procedure SetShowGridLines(const Value : Boolean);
  protected
    function CellAtPos(const X, Y : Longint) : TPoint;
    function CellCoords(const X : TGridWidthRange; const Y : TGridHeightRange) : TRect;
    procedure Change; dynamic;
    function DoesCellLive(const X : TGridWidthRange; const Y : TGridHeightRange) : Boolean; dynamic;
    procedure InvalidateCell(const X : TGridWidthRange; const Y : TGridHeightRange);
    procedure MouseUp(Button : TMouseButton; Shift : TShiftState; X, Y : Integer); override;
    function NumberOfNeighbours(const X : TGridWidthRange; const Y : TGridHeightRange) : Integer; dynamic;
    procedure Paint; override;
    procedure SetGridSize(const NewGridWidth : TGridWidthRange;
                          const NewGridHeight : TGridHeightRange);
  public
    constructor Create(aOwner : TComponent); override;
    procedure ClearCells;
    procedure NextGeneration;
    destructor Destroy; override;
    property Cells[const X : TGridWidthRange; const Y : TGridHeightRange] : Boolean
             read GetCells write SetCells;
    property Generation : Cardinal read FGeneration;
    property LiveCells : Integer read GetLiveCells;
    property Version : String read GetVersion;
  published
    property AcceptMouseClicks : Boolean read FAcceptMouseClicks write SetAcceptMouseClicks
             default False;
    property Align;
    property CellColor : TColor read FCellColor write SetCellColor default DefaultCellColor;
    property Color;
    property DragCursor;
    property DragMode;
    property GridHeight : TPermissableGridHeight read FGridHeight write SetGridHeight
             default DefaultGridHeight;
    property GridLineColor : TColor read FGridLineColor write SetGridLineColor
             default DefaultGridLineColor;
    property GridLineStyle : TPenStyle read FGridLineStyle write SetGridLineStyle
             default DefaultGridLineStyle;
    property GridWidth : TPermissableGridWidth read FGridWidth write SetGridWidth
             default DefaultGridWidth;
    property OnChange : TNotifyEvent read FOnChange write FOnChange;
    property ParentColor;
    property PopupMenu;
    property ShowHint;
    property ShowGridLines : Boolean read FShowGridLines write SetShowGridLines
             default False;
    property Visible;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
{$ifndef VER80}
    property OnStartDrag;
{$endif}
  end;

procedure Register;

implementation

const
  CurrentVersion = '1.0';

{ Returns the cell containing the specified pixel position }
function TLife.CellAtPos(const X, Y : Longint) : TPoint;

{ I AM STILL SUSPICIOUS ABOUT THIS }
begin
{$IFOPT R+}
  if X >= ClientWidth then
    raise ERangeError.Create('X parameter out of range');
  if Y >= ClientHeight then
    raise ERangeError.Create('Y parameter out of range');
{$ENDIF}
  Result.X := X div (ClientWidth div GridWidth);
  Result.Y := Y div (ClientHeight div GridWidth)
end;

{ Returns the bounds of the specified cell }
function TLife.CellCoords(const X : TGridWidthRange; const Y : TGridHeightRange) : TRect;

begin
{$IFOPT R+}
  if X >= GridWidth then
    raise ERangeError.Create('X parameter out of range');
  if Y >= GridHeight then
    raise ERangeError.Create('Y parameter out of range');
{$ENDIF}
  with Result do
  begin
    Left := ClientWidth * x div GridWidth;
    Top := ClientHeight * y div GridHeight;
    Right := ClientWidth * (x + 1) div GridWidth;
    Bottom := ClientHeight * (y + 1) div GridHeight
  end
end;

{ Returns the offset in Grid (which is only 1-dimensional) of the specified cell
  (which are on a 2-dimensional plane). }
function TLife.CellCoordToGridOffset(const X : TGridWidthRange; const Y : TGridHeightRange) : Cardinal;

begin
{$IFOPT R+}
  if X >= GridWidth then
    raise ERangeError.Create('X parameter out of range');
  if Y >= GridHeight then
    raise ERangeError.Create('Y parameter out of range');
{$ENDIF}
  Result := Y * GridWidth + X
end;

{ Call the OnChange event. }
procedure TLife.Change;

begin
  if Assigned(FOnChange) then
    FOnChange(Self)
end;

{ Mark all the cells as "dead" }
procedure TLife.ClearCells;

begin
  FillChar(FGrid^, GridWidth * GridHeight, 0);
  Invalidate;
  Change
end;

constructor TLife.Create(aOwner : TComponent);

begin
  Inherited Create(aOwner);
  SetGridSize(DefaultGridHeight, DefaultGridWidth);
  FCellColor := DefaultCellColor;
  FGridLineColor := DefaultGridLineColor;
  FGridLineStyle := DefaultGridLineStyle;
  Height := 50;
  Width := 50
end;

destructor TLife.Destroy;

begin
  if FGrid <> nil then
{$IFDEF VER80}
    FreeMem(FGrid, FGridSize);
{$ELSE}
    FreeMem(FGrid);
{$ENDIF}
  Inherited Destroy
end;

function TLife.DoesCellLive(const X : TGridWidthRange; const Y : TGridHeightRange) : Boolean;

begin
{$IFOPT R+}
  if X >= GridWidth then
    raise ERangeError.Create('X parameter out of range');
  if Y >= GridHeight then
    raise ERangeError.Create('Y parameter out of range');
{$ENDIF}
  case NumberOfNeighbours(X, Y) of
    2    : Result := Cells[X, Y];
    3    : Result := True;
    else   Result := False
  end
end;

{ Is the cell at the specified position alive? }
function TLife.GetCells(const X : TGridWidthRange; const Y : TGridHeightRange) : Boolean;

begin
{$IFOPT R+}
  if X >= GridWidth then
    raise ERangeError.Create('X parameter out of range');
  if Y >= GridHeight then
    raise ERangeError.Create('Y parameter out of range');
{$ENDIF}
  Result := Boolean(FGrid^[CellCoordToGridOffset(X, Y)])
end;

{ Returns the number of "live" cells }
function TLife.GetLiveCells : Integer;

var
  x : Integer;
  y : Integer;

begin
  Result := 0;
  for x := 0 to GridWidth - 1 do
    for y := 0 to GridHeight - 1 do
      if Cells[x, y] then
        Inc(Result)
end;

function TLife.GetVersion : String;

begin
  Result := CurrentVersion
end;

procedure TLife.InvalidateCell(const X : TGridWidthRange; const Y : TGridHeightRange);

var
  Rect : TRect;

begin
{$IFOPT R+}
  if X >= GridWidth then
    raise ERangeError.Create('X parameter out of range');
  if Y >= GridHeight then
    raise ERangeError.Create('Y parameter out of range');
{$ENDIF}
  Rect := CellCoords(X, Y);
  InvalidateRect(Handle, @Rect, True)
end;

procedure TLife.MouseUp(Button : TMouseButton; Shift : TShiftState; X, Y : Integer);

var
  ClickedCell : TPoint;

begin
  if AcceptMouseClicks and (Button = mbLeft) then
  begin
    ClickedCell := CellAtPos(X, Y);
    Cells[ClickedCell.X, ClickedCell.Y] := not Cells[ClickedCell.X, ClickedCell.Y]
  end
end;

procedure TLife.NextGeneration;

var
  x : Integer;
  y : Integer;

begin 
  for x := 0 to GridWidth - 1 do
    for y := 0 to GridHeight - 1 do
      SetCell(x, y, DoesCellLive(x, y));
  Inc(FGeneration);
  Change
end;

function TLife.NumberOfNeighbours(const X : TGridWidthRange; const Y : TGridHeightRange) : Integer;

var
  xMinus1 : Integer;
  xPlus1  : Integer;
  yMinus1 : Integer;
  yPlus1  : Integer;

begin
{$IFOPT R+}
  if X >= GridWidth then
    raise ERangeError.Create('X parameter out of range');
  if Y >= GridHeight then
    raise ERangeError.Create('Y parameter out of range');
{$ENDIF}
  Result := 0;
  xMinus1 := X - 1;
  if xMinus1 < 0 then
    xMinus1 := GridWidth - 1;
  xPlus1 := (X + 1) mod GridWidth;
  yMinus1 := Y - 1;
  if yMinus1 < 0 then
    yMinus1 := GridHeight - 1;
  yPlus1 := (Y + 1) mod GridHeight;
  if Cells[xMinus1, yMinus1] then
    Inc(Result);
  if Cells[X, yMinus1] then
    Inc(Result);
  if Cells[xPlus1, yMinus1] then
    Inc(Result);
  if Cells[xMinus1, Y] then
    Inc(Result);
  if Cells[xPlus1, Y] then
    Inc(Result);
  if Cells[xMinus1, yPlus1] then
    Inc(Result);
  if Cells[X, yPlus1] then
    Inc(Result);
  if Cells[xPlus1, yPlus1] then
    Inc(Result)
end;

procedure TLife.Paint;

var
  x : Integer;
  y : Integer;
  r : TRect;

begin
  with Canvas do
  begin
    { Draw the grid lines }
    if ShowGridLines then
    begin
      Brush.Color := Color;
      Brush.Style := bsSolid;
      Pen.Color := GridLineColor;
      Pen.Mode := pmMask;
      Pen.Style := GridLineStyle;
      for x := 1 to GridWidth - 1 do
      begin
        MoveTo(ClientWidth * x div GridWidth, 0);
        LineTo(ClientWidth * x div GridWidth, ClientHeight)
      end;
      for y := 1 to GridHeight - 1 do
      begin
        MoveTo(0, ClientHeight * y div GridHeight);
        LineTo(ClientWidth, ClientHeight * y div GridHeight)
      end;
    end;
    { Draw all the live cells }
    Pen.Style := psClear;
    Brush.Color := CellColor;
    Brush.Style := bsSolid;
    for x := 0 to GridWidth - 1 do
      for y := 0 to GridHeight - 1 do
        if Cells[x, y] then
        begin
          r := CellCoords(x, y);
          with r do
            Ellipse(Left + 2, Top + 2, Right - 1, Bottom - 1)
        end;
    { At design-time, draw a dashed line around the component }
    if csDesigning in ComponentState then
    begin
      Brush.Color := Color;
      Brush.Style := bsCross;
      FrameRect(Rect(0, 0, Width, Height))
    end
  end
end;

procedure TLife.SetAcceptMouseClicks(const Value : Boolean);

begin
  if Value <> FAcceptMouseClicks then
  begin
    FAcceptMouseClicks := Value;
    Change
  end
end;

procedure TLife.SetCell(const X : TGridWidthRange; const Y : TGridHeightRange; const Value : Boolean);

begin
{$IFOPT R+}
  if X >= GridWidth then
    raise ERangeError.Create('X parameter out of range');
  if Y >= GridHeight then
    raise ERangeError.Create('Y parameter out of range');
{$ENDIF}
  if Cells[X, Y] <> Value then
  begin
    FGrid^[CellCoordToGridOffset(X, Y)] := Byte(Value);
    InvalidateCell(X, Y)
  end
end;

procedure TLife.SetCellColor(Const Value : TColor);

begin
  if Value <> FCellColor then
  begin
    FCellColor := Value;
    Invalidate
  end
end;

procedure TLife.SetCells(const X : TGridWidthRange; const Y : TGridHeightRange; const Value : Boolean);

begin
  if Cells[X, Y] <> Value then
  begin
    SetCell(X, Y, Value);
    FGeneration := 0;
    Change
  end
end;

procedure TLife.SetGridHeight(const NewGridHeight : TPermissableGridHeight);

begin
  SetGridSize(GridWidth, NewGridHeight)
end;

procedure TLife.SetGridLineColor(Const Value : TColor);

begin
  if Value <> FGridLineColor then
  begin
    FGridLineColor := Value;
    Invalidate
  end
end;

procedure TLife.SetGridLineStyle(Const Value : TPenStyle);

begin
  if Value <> FGridLineStyle then
  begin
    FGridLineStyle := Value;
    Invalidate
  end
end;

procedure TLife.SetGridSize(const NewGridWidth : TGridWidthRange;
                            const NewGridHeight : TGridHeightRange);

begin
  if (NewGridWidth <> GridWidth) or (NewGridHeight <> GridHeight) then
  begin
    FGridWidth := NewGridWidth;
    FGridHeight := NewGridHeight;
    { Free the old grid }
{$IFDEF VER80}
    FreeMem(FGrid, FGridSize);
{$ELSE}
    FreeMem(FGrid);
{$ENDIF}
    { Allocate a new grid }
    try
      FGrid := AllocMem(NewGridWidth * NewGridHeight);
{$IFDEF VER80}
      FGridSize := NewGridWidth * NewGridHeight
{$ENDIF}
    except
      FGrid := nil;
      raise
    end;
    { Redraw the component }
    Invalidate;
    { Let the OnChange event know about the changes }
    Change
  end
end;

procedure TLife.SetGridWidth(const NewGridWidth : TPermissableGridWidth);

begin
  SetGridSize(NewGridWidth, GridHeight)
end;

procedure TLife.SetShowGridLines(const Value : Boolean);

begin
  if Value <> FShowGridLines then
  begin
    FShowGridLines := Value;
    Invalidate
  end
end;

procedure Register;

begin
  RegisterComponents('Extras', [TLife])
end;

end.
